/*******************************************************************************

  Defines Filter Interface Utils for DIMIN Viewer n5 (DIV5)

  Programmer: Dima V. Fedorov Levit <dimin@dimin.net> <http://www.dimin.net/>

  History:
    06/23/2004 23:37 - First Creation

*******************************************************************************/

#include "DIVFilterUtils.h"


DIM_UCHAR iTrimUC (int num)
{
  if (num < 0) return 0;
  if (num > 255) return 255;
  return num;
}


//******************************************************************************
// Safe call to callbacks
//******************************************************************************

void showProgress( TDIVParameters *par, long done, long total, char *descr )
{
  if (par == NULL) return;
  if (par->progressProc == NULL) return;

  par->progressProc( done, total, descr );
}

int testAbort( TDIVParameters *par )
{
  if (par == NULL) return 0;
  if (par->testAbortProc == NULL) return 0;

  return par->testAbortProc();
}

void showError( TDIVParameters *par, int val, char *descr )
{
  if (par == NULL) return;
  if (par->errorProc == NULL) return;

  par->errorProc( val, descr );
}

//******************************************************************************
// Safe way to find out selection line constrains
//******************************************************************************

void findLineBegEndTopBotI(TDIVParameters *par, uint w, long y, uint &pI, int& pBeg, int& pEnd)
{
  uint pBegU = pBeg;
  uint pEndU = pEnd;
  findLineBegEndTopBot(par, w, y, pI, pBegU, pEndU);
  pBeg = pBegU;
  pEnd = pEndU;
}

// w - image width, y - line for which to get constrains
// pBeg and pEnd are retrieved begining and ending points
void findLineBegEndTopBot(TDIVParameters *par, uint w, long y, uint &pI, uint& pBeg, uint& pEnd)
{
  pBeg = 0;
  pEnd = w-1;

  #if defined(__BORLANDC__)
  DIV5::TPoint *chain = par->selection.chain;
  #else
  TPoint *chain = par->selection.chain;
  #endif

  
  if (par == NULL) return;
  if (par->selection.chainCount <= 0) return;  
  if (par->selection.chain == NULL) return; 

  pBeg = 1;
  pEnd = 0;
  if (pI >= par->selection.chainCount) return;

  //find first
  while ( ( (*(chain+pI)).y < y) && (pI<par->selection.chainCount) ) { pI++; }
  if ( (*(chain+pI)).y == y) pBeg = (*(chain+pI)).x; 

  //find last
  while ( (chain[pI].y == y) && (pI<par->selection.chainCount) ) pI++;
  if ( (pI>0) && (chain[pI-1].y == y) ) pEnd = chain[pI-1].x;

}


//******************************************************************************
// @@@ ARGUMENTS
//******************************************************************************

bool getUIntArg   (TDIVParameters *par, uint i, uint &arg )
{
  if (par == NULL) return FALSE;
  if (par->args == NULL) return FALSE;
  if (par->args->count <= i) return FALSE;
  if (par->args->args == NULL) return FALSE;
  if (par->args->args[i].tagCount <= 0) return FALSE;
  if ((par->args->args[i].tagType != DIM_TAG_LONG) &&
      (par->args->args[i].tagType != DIM_TAG_ENUM) ) return FALSE;

  arg = par->args->args[i].tagData.uVal;
  return TRUE;
}

bool getIntArg    (TDIVParameters *par, uint i, int &arg )
{
  if (par == NULL) return FALSE;
  if (par->args == NULL) return FALSE;
  if (par->args->count <= i) return FALSE;
  if (par->args->args == NULL) return FALSE;
  if (par->args->args[i].tagCount <= 0) return FALSE;
  if (par->args->args[i].tagType != DIM_TAG_LONG) return FALSE;

  arg = par->args->args[i].tagData.iVal;
  return TRUE;
}

bool getByteArg   (TDIVParameters *par, uint i, uchar &arg )
{
  if (par == NULL) return FALSE;
  if (par->args == NULL) return FALSE;
  if (par->args->count <= i) return FALSE;
  if (par->args->args == NULL) return FALSE;
  if (par->args->args[i].tagCount <= 0) return FALSE;
  if (par->args->args[i].tagType != DIM_TAG_BYTE) return FALSE;

  arg = (uchar) par->args->args[i].tagData.pVal;
  return TRUE;
}

bool getFloatArg  (TDIVParameters *par, uint i, float &arg )
{
  if (par == NULL) return FALSE;
  if (par->args == NULL) return FALSE;
  if (par->args->count <= i) return FALSE;
  if (par->args->args == NULL) return FALSE;
  if (par->args->args[i].tagCount <= 0) return FALSE;
  if (par->args->args[i].tagType != DIM_TAG_FLOAT) return FALSE;

  arg = par->args->args[i].tagData.fVal;
  return TRUE;
}

bool getDoubleArg (TDIVParameters *par, uint i, double &arg )
{
  if (par == NULL) return FALSE;
  if (par->args == NULL) return FALSE;
  if (par->args->count <= i) return FALSE;
  if (par->args->args == NULL) return FALSE;
  if (par->args->args[i].tagCount <= 0) return FALSE;
  if (par->args->args[i].tagData.pVal == NULL) return FALSE;
  if (par->args->args[i].tagType != DIM_TAG_RATIONAL) return FALSE;

  double *pd;
  pd = (double *) par->args->args[i].tagData.pVal; 
  arg = *pd;
  return TRUE;  
}

bool getPointerArg (TDIVParameters *par, uint i, void **arg )
{
  if (par == NULL) return FALSE;
  if (par->args == NULL) return FALSE;
  if (par->args->count <= i) return FALSE;
  if (par->args->args == NULL) return FALSE;
  if (par->args->args[i].tagCount <= 0) return FALSE;
  if (par->args->args[i].tagData.pVal == NULL) return FALSE;
  if (divTypeSizeBytes[par->args->args[i].tagType] < divTypeSizeBytes[DIM_TAG_POINTER]) return FALSE;

  *arg = par->args->args[i].tagData.pVal;
  return TRUE;
}

bool getBoolArg   (TDIVParameters *par, uint i, bool &arg )
{
  if (par == NULL) return FALSE;
  if (par->args == NULL) return FALSE;
  if (par->args->count <= i) return FALSE;
  if (par->args->args == NULL) return FALSE;
  if (par->args->args[i].tagCount <= 0) return FALSE;
  if (par->args->args[i].tagType != DIM_TAG_BOOL) return FALSE;

  #pragma warning( disable : 4800 )
  arg = (bool) par->args->args[i].tagData.iVal;
  #pragma warning( default : 4800 )
  return TRUE;
}


//******************************************************************************
// @@@ ARGUMENT LIST CREATION
//******************************************************************************

TDIVArguments createFilterArguments(int count )
{
  TDIVArguments args;
  args.usePreview = 0;
  args.count = 0;  

  if (count <= 0) return args;

  args.args = (TDIVTagedArgument *) new uchar [ sizeof(TDIVTagedArgument) * count ];
  if (args.args != NULL) {
    args.usePreview = 1;
    args.count = count;
  }
  return args;
}

void freeFilterArguments( TDIVArguments *args )
{
  if (args == NULL) return;
  if ( (args->count > 0) && (args->args != NULL) )
    delete args->args;
  args->usePreview = 0;
  args->count = 0;  
}

void changeFilterArgument( TDIVArguments *args, uint index,
                           DWORD tagId, DWORD tagType, DWORD tagCount,
                           char *tagDescr, void *tagData )
{
  if (args == NULL) return;
  if ( (args->count <= index) || (args->args == NULL) ) return; 
  
  args->args[ index ].tagId    = tagId;
  args->args[ index ].tagType  = tagType;
  args->args[ index ].tagCount = tagCount;
  args->args[ index ].tagDescr = tagDescr;
  args->args[ index ].tagData.pVal  = tagData;
}


//------------------------------------------------------------------------------
// PALETTE
//------------------------------------------------------------------------------

HPALETTE createGreyPalette( )
{
  HPALETTE hpal;
  LOGPALETTE *pal;
  int i;
  int palSize; //Size of the palette buffer

  palSize = sizeof(LOGPALETTE) + (sizeof(PALETTEENTRY) * 255);
  pal = (LOGPALETTE *) malloc(palSize);

  pal->palVersion = 0x300;
  pal->palNumEntries = 256;

  // Set the colors to their Gray indexes
  for (i = 0 ; i <= 255; i++)
  {
    pal->palPalEntry[i].peRed   = (unsigned char) i;
    pal->palPalEntry[i].peGreen = (unsigned char) i;
    pal->palPalEntry[i].peBlue  = (unsigned char) i;
    pal->palPalEntry[i].peFlags = PC_NOCOLLAPSE;
  }

  hpal = CreatePalette(pal);
  free(pal);
  return hpal;
}

HPALETTE createHPalette( int num_colors, RGBQUAD *bmiColors )
{
  HPALETTE hpal;
  LOGPALETTE *pal;
  int i;
  int palSize; //Size of the palette buffer

  palSize = sizeof(LOGPALETTE) + (sizeof(PALETTEENTRY) * (num_colors-1) );
  pal = (LOGPALETTE *) malloc(palSize);

  pal->palVersion = 0x300;
  pal->palNumEntries = num_colors;

  // Set the colors to their Gray indexes
  for (i = 0 ; i < num_colors; i++)
  {
    pal->palPalEntry[i].peRed   = (unsigned char) bmiColors[i].rgbRed;
    pal->palPalEntry[i].peGreen = (unsigned char) bmiColors[i].rgbGreen;
    pal->palPalEntry[i].peBlue  = (unsigned char) bmiColors[i].rgbBlue;
    pal->palPalEntry[i].peFlags = PC_NOCOLLAPSE;
  }

  hpal = CreatePalette(pal);
  free(pal);
  return hpal;
}

//------------------------------------------------------------------------------
// image
//------------------------------------------------------------------------------

DIM_UINT getHBLineSizeInBytes( DIM_UINT w, DIM_UINT depth )
{
  return (((w * depth +31) & ~31)/8);
}

DIM_UCHAR *getHBScanLine( DIM_UINT y, DIM_UCHAR *p, DIM_UINT w, DIM_UINT depth )
{
  DIM_UINT ls = getHBLineSizeInBytes( w, depth );
  return p + (ls * y);
}


HBITMAP dimCreateHBitmap( DIM_UINT w, DIM_UINT h, DIM_UINT depth, int num_colors, RGBQUAD *lut, DIM_UCHAR **p, HPALETTE &hpal )
{
  int bpp = depth;
  HBITMAP hbmp;

  // Allocate memory for the BITMAPINFO structure. (This structure 
  // contains a BITMAPINFOHEADER structure and an array of RGBQUAD 
  // data structures.) 

  BITMAPINFO *pbmi;
  DIM_UINT pbmi_size = 0;

  if (bpp != 24) 
    pbmi_size = sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD) * (1<< bpp); 
  else 
    // There is no RGBQUAD array for the 24-bit-per-pixel format. 
    pbmi_size = sizeof(BITMAPINFOHEADER);

  pbmi = (PBITMAPINFO) LocalAlloc(LPTR, pbmi_size ); 
  memset(pbmi, 0, pbmi_size );


  // Initialize the fields in the BITMAPINFO structure. 
  pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER); 
  pbmi->bmiHeader.biWidth = w; 
  pbmi->bmiHeader.biHeight = h; 
  pbmi->bmiHeader.biPlanes = 1; 
  pbmi->bmiHeader.biBitCount = bpp; 
  if (bpp < 24) pbmi->bmiHeader.biClrUsed = (1<<bpp); 

  // If the bitmap is not compressed, set the BI_RGB flag. 
  pbmi->bmiHeader.biCompression = BI_RGB; 

  // Compute the number of bytes in the array of color 
  // indices and store the result in biSizeImage. 
  // For Windows NT, the width must be DWORD aligned unless 
  // the bitmap is RLE compressed. This example shows this. 
  // For Windows 95/98/Me, the width must be WORD aligned unless the 
  // bitmap is RLE compressed.
  pbmi->bmiHeader.biSizeImage = ((w * bpp +31) & ~31) /8 * h; 
  // Set biClrImportant to 0, indicating that all of the 
  // device colors are important. 
  pbmi->bmiHeader.biClrImportant = 0; 

  if ( (num_colors > 0) && (lut != NULL) )
  {
    int i;
    for (i=0; i<num_colors; i++)
    {
      pbmi->bmiColors[i].rgbRed   = lut[i].rgbRed;
      pbmi->bmiColors[i].rgbGreen = lut[i].rgbGreen;
      pbmi->bmiColors[i].rgbBlue  = lut[i].rgbBlue;
    }
    hpal = createHPalette( 256, pbmi->bmiColors );
  }

  hbmp = CreateDIBSection(NULL, pbmi, DIB_RGB_COLORS, (void**)p, NULL, 0);
  GdiFlush();
  
  if (p == NULL) return 0;

  GdiFlush();
  return hbmp;
}

HBITMAP dimCopyHBitmap( DIM_UCHAR *src, DIM_UINT w, DIM_UINT h, DIM_UINT depth, int num_colors, RGBQUAD *lut, DIM_UCHAR **p )
{
  HPALETTE hpal;
  HBITMAP res = dimCreateHBitmap( w, h, depth, num_colors, lut, p, hpal );
  memcpy( *p, src, getHBLineSizeInBytes(w, depth) * h );
  return res;
}







